/* userdb.c
**
** Database functions for vxmail
*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include "prefs.h"
#include "dbif.h"
#include "vxerror.h"
#include "tools.h"
#include "sysauth.h"
#include "userdb.h"
#include "inline-funcs.h"
#include "mrmacs.h"

#define MAXTRIES 10
#define TRYDELAY 5

static DBIF_FILE userdb = (DBIF_FILE) 0;

/* local function for initializing the userdb */
void init_userdb (void)
{
    int tries = 0;
    if (userdb) { return; }

    /* try MAXTRIES times to establish a database connection */
    for (;;) {
	dbif_reset (&userdb, prefs.userdb_name);
	if (userdb || tries++ >= MAXTRIES) { break; }
	sleep (TRYDELAY);
    }

    /* if the connection failed, abort the program execution */
    if (!userdb) {
	vxerror_at (EX_NOINPUT, errno, __FILE__, __LINE__ - 7,
		    "user database `%s'", prefs.userdb_name);
    }
}

/* check for the existance of an e-mail domain */
int domain_check (char *em_addr)
{
    dbif_rec *rec;
    char *domain = strchr (em_addr, '@');
    if (domain) ++domain; else domain = em_addr;
    init_userdb ();
    rec = dbif_get (userdb, domain);
    if (rec) {
	int rt = dbif_rectype(rec);
	dbif_free (rec);
	return (rt == RT_DOMAIN);
    }
    return 0;
}

int xdomain_check (char *em_addr)
{
    dbif_rec *rec;
    char *domain = strchr (em_addr, '@');
    if (domain) ++domain; else domain = em_addr;
    init_userdb ();
    rec = dbif_get (userdb, domain);
    if (rec) {
	int rt = dbif_rectype(rec);
	dbif_free (rec);
	if (rt != RT_DOMAIN && rt != RT_NSDOMAIN) {
	    return 0;
	}
	return (rt == RT_DOMAIN ? 1 : -1);
    }
    return 0;
}

/* make a copy of an e-mail address and convert this to lower case
** (if required).
**
** I implemented this function because i found some problems in conjunction
** with BSMTP folders and a client of the company which uses VxMail for
** local delivery. With this function, i will reduce the conversion problem
** to the only place where a conversion is required - in database queries.
*/
char *mkconvaddr (char *em_addr)
{
    char *new_addr = gdupl (__FILE__, __LINE__, em_addr);
    char *p = strchr (new_addr, '@');
    ifnull (p) { p = new_addr + strlen (new_addr); }
    if (prefs.lcuser) {
	if (strncmp (new_addr, "::", 2)
	&&  strncmp (new_addr, "<OTHER>", 7)
	&&  strncmp (new_addr, "<PRE>", 5)) {
	    lcconv (new_addr, p - new_addr);
	}
    }
    if (prefs.lcdomain) { lcconv (p, -1); }
    return new_addr;
}

/* get_ffile (addr)
**     returns the name of a filter file for the domain, the user `addr' is
**     a member of - or NULL (if the domain doesn't have a filter).
*/
char *get_ffile (char *em_addr)
{
    dbif_rec *rec;
    char *p, *faddr, *ffile;
    ifnull (p = strchr (em_addr, '@')) { return NULL; }
    faddr = gconc (__FILE__, __LINE__, "::FILTER::", p, NULL);
    if (prefs.lcdomain) { lcconv (&faddr[11], -1); }
    ifnull (rec = dbif_get (userdb, faddr)) { gfree (faddr); return NULL; }
    if (dbif_rectype(rec) != RT_FILTER) {
	dbif_free (rec); gfree (faddr);
	return NULL;
    }
    ffile = gdupl (__FILE__, __LINE__, rec->folder);
    dbif_free (rec); gfree (faddr);
    return ffile;
}

/* get a user's folder;
** if the user entry is a link, then resolve this link first
*/

char *user_folder (char *_em_addr, size_t *ml, off_t *fl, int *rtype)
{
    dbif_rec *rec;
    char *wfolder, *em_addr;
    size_t fll;
    int rt;
    char *p, *q;
    init_userdb ();
    em_addr = mkconvaddr (_em_addr);
    *ml = 0; *fl = 0;
    if (strcmp (em_addr, "::MAILSYS::")) {
	p = strchr (em_addr, '@');
	if (isNULL(p) || !*++p) { return (char *) 0; }
	/* (DIRTY!!!) Workaround for the #!?$&!$& Claris mailer */
	if ((q = strchr (em_addr, ':'))) {
	    if ((p = strchr (em_addr, '@')) && (p - q > 0)) {
		*q = '@'; *p = '\0';
	    }
	}
    }
    rec = dbif_resolve (userdb, em_addr, NULL, prefs.iu_receives);
    ifnull (rec) {
	char *alladdr;
	if (!strcmp (em_addr, "::MAILSYS::")) { gfree (em_addr); return NULL; }
	if (!strncmp (em_addr, "<PRE>@", 6)) { gfree (em_addr); return NULL; }
	p = strchr (em_addr, '@');
	alladdr = gconc (__FILE__, __LINE__, "<OTHER>", p, NULL);
	p = strchr (alladdr, '@');
	rec = dbif_resolve (userdb, alladdr, NULL, prefs.iu_receives);
	gfree (alladdr);
	ifnull (rec) { gfree (em_addr); return NULL; }
    }
    rt = dbif_rectype(rec);
    if (!((1<<rt)&(CC_USER|CC_NSUSER|CC_ADMIN|CC_MAILSYS|CC_LLINK|CC_BSMTP))) {
	vxlog (1, "  wrong user type!\n");
	dbif_free (rec);
	gfree (em_addr);
	return NULL;
    }
    fll = strlen (rec->folder);
    if (rt == RT_BSMTP) {
	wfolder = gdupl (__FILE__, __LINE__, rec->folder);
    } else if ((1 << rt) & (CC_RUSERS|CC_MAILSYS)) {
	wfolder = abs_path (prefs.mbox_base, rec->folder);
    } else if (rt == RT_FILTER) {
	wfolder = abs_path (prefs.listdir, rec->folder);
    } else if (rt == RT_LLINK) {
	char *p = rec->folder;
	size_t fll = 1, sl;
	int mx = ((int) *p++) + 1, ix;
	for (ix = 0; ix < mx; ++ix) { sl = strlen (p) + 1; fll += sl; p += sl; }
	wfolder = galloc (__FILE__, __LINE__, fll);
	memcpy (wfolder, rec->folder, fll);
    } else {
	wfolder = gdupl (__FILE__, __LINE__, rec->folder);
    }
    *ml = rec->msg_limit;
    *fl = rec->fld_limit;
    *rtype = dbif_rectype (rec);
    dbif_free (rec);
    gfree (em_addr);
    return wfolder;
}

char *user_rewrite (char *_em_addr, size_t *ml, off_t *fl, int *rtype)
{
    dbif_rec *rec;
    char *new_user, *dompart, *wfolder, *domname, *em_addr;
    size_t nulen, uplen, fll;
    int rt;
    init_userdb ();
    *ml = 0; *fl = 0;
    em_addr = mkconvaddr (_em_addr);
    dompart = strchr (em_addr, '@');
    /* `sendmail' doesn't set the `$h' macro in the case of the local host. */ 
    if (isNULL(dompart) || !dompart[1]) {
	/* configurable domainname for the local host! */
	uplen = (dompart ? (dompart - em_addr) : strlen (em_addr));
	dompart = mailhost ();
	vxlog (5, "no domain found - using `%s' instead!\n", dompart);
    } else {
	uplen = dompart - em_addr;
	++dompart;
    }
    rec = dbif_resolve (userdb, dompart, &domname, prefs.iu_receives);
    ifnull (rec) { gfree (em_addr); return NULL; }
    rt = dbif_rectype(rec);
    /* The result of the database access MUST be a domain entry! */
    if (rt != RT_DOMAIN && rt != RT_NSDOMAIN) {
	dbif_free (rec);
	gfree (em_addr);
	return NULL;
    }
    fll = strlen (domname);
    nulen = uplen + fll + 1;
    /* allocate enough memory for the new e-mail address */
    new_user = galloc (__FILE__, __LINE__, nulen + 1);
    /* concatenate the user part from the old e-mail address and the
    ** domain part (from the `folder' element of the domain link record
    ** to a new e-mail address
    */
    memcpy (new_user, em_addr, uplen);
    new_user[uplen] = '@';
    memcpy (new_user + uplen + 1, domname, fll);
    new_user[nulen] = '\0';
    vxlog (1, "retrying as `%s' ...", new_user);
    /* get the folder path for the rewritten e-mail address */
    wfolder = user_folder (new_user, ml, fl, rtype);
    /* If the result is empty, check if the domain is a non-standard domain.
    ** If so, try again the `sysauth'-mechanism.
    */
    if (isNULL(wfolder)
    &&  ((1 << dbif_rectype(rec)) & CC_TRUEDOM)
    &&  prefs.sysauth) {
	dbif_rec *rec1;
	vxlog (1, " (with system authentication)");
	rec1 = sysauth (rec, em_addr);
	wfolder = NULL;
	if (rec1) {
	    wfolder = abs_path (prefs.mbox_base, rec1->folder);
	    *ml = 0; *fl = 0;
	    *rtype = dbif_rectype(rec1);
	    dbif_free (rec1);
	}
    }
    /* release the memory which was allocated for the domain link record */
    dbif_free (rec);
    free (domname);
    /* release the memory intermediately allocated for the rewritten
    ** e-mail address
    */
    gfree (new_user);
    vxlog (1, " %s.\n", wfolder ? "done" : "failed");
    gfree (em_addr);
    return wfolder;
}

static dbif_rec *idrec = (dbif_rec *) 0;

char *mailhost (void)
{
    static char *mh = NULL;
    dbif_rec *rec;
    if (mh) return mh;
    init_userdb ();
    ifnull (idrec) { idrec = dbif_get (userdb, "::VXMAIL::"); }
    rec = idrec;
    ifnull (rec) {
	vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__,
		    "database inconsistency! "
		    "This is a FATAL ERROR! Exiting!\n");
    }
    mh = gdupl (__FILE__, __LINE__, rec->passwd);
    return mh;
}

char *defadmin (void)
{
    static char *da = NULL;
    dbif_rec *rec;
    unlessnull (da) { return da; }
    init_userdb ();
    ifnull (idrec) { idrec = dbif_get (userdb, "::VXMAIL::"); }
    rec = idrec;
    ifnull (rec) {
	vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__,
		    "database inconsistency! "
		    "This is a FATAL ERROR! Exiting!\n");
    }
    da = gdupl (__FILE__, __LINE__, rec->folder);
    return da;
}
